#include "test/Service.h"
#include "test/Tester.h"

#include <boost/algorithm/string/join.hpp>
#include <boost/format.hpp>
#include <boost/range/adaptors.hpp>
#include <nlohmann/json.hpp>

#include <chrono>
#include <functional>
#include <thread>


namespace xs { namespace test {

namespace detail {
    std::string toString(const nlohmann::json& object)
    {
        return object.is_string() ? object.get<std::string>()
                                  : object.dump();
    }
}

Tester::Tester(const boost::filesystem::path& file)
: config_(file)
, server_(
    "xs-test", { }, { }, 0, nullptr
)
{
    registerServices();
}

void Tester::registerServices() noexcept
{
    using namespace std::placeholders;
    for (const auto& service : config_.getServices()) {
        if (service.isCommand()) {
            server_.onCmd(
                service.name_,
                std::bind(&Tester::onCommand, this, _1, _2)
            );
        } else {
            server_.onEvent(
                service.name_,
                std::bind(&Tester::onEvent, this, _1, _2)
            );
        }
    }
} 

void Tester::resetCounters() noexcept
{
    receivedCommandCount_.clear();
    receivedEventCount_.clear();
}

Tester::Report Tester::createReport(const std::string& response, const Expectation& expectation) noexcept
{
    Report report{ true };
    // execute withou expected response
    if (!expectation.response_) {
        return report;
    }

    if (expectation.response_->is_string()) {
         report.pass = (response == expectation.response_->get<std::string>());
    } else {
         try {
             report.pass = (nlohmann::json::parse(response) == *expectation.response_);
         } catch (...) {

         }
    }

    if (!expectation.runs_.empty()) {
        for (const auto& command : expectation.runs_) {
            if (!receivedCommandCount_[command]) {
                report.missedCommands.emplace(command);
                report.pass = false;
            }
        }
    }

    if (!expectation.posts_.empty()) {
        for (const auto& event : expectation.posts_) {
            if (!receivedEventCount_[event]) {
                report.missedEvents.emplace(event);
                report.pass = false;
            }
        }
    }

    return report;
}

std::string Tester::onCommand(const std::string& name, const std::string& params) noexcept
{
    using namespace boost::adaptors;
    using namespace std::placeholders;
    const auto handlers = config_.getServices()
                        | filtered(std::bind(&Service::isCommand, _1));
    for (const auto& handler : handlers) {
        if (handler.name_ != name) {
            continue;
        }

        if (!handler.request_) {
            receivedCommandCount_[name]++;
            return detail::toString(*handler.response_);
        }

        auto& receiveCount = receivedCommandCount_[name];

        const auto& request = *handler.request_;
        const bool isString = request.is_string();
        if (isString) {
            if (params == request.get<std::string>()) {
                receiveCount++;
                return detail::toString(*handler.response_);
            }
        } else {
            if (nlohmann::json::parse(params) == request) {
                receiveCount++;
                return detail::toString(*handler.response_);
            }
        }
    }

    return "";
}

void Tester::onEvent(const std::string& name, const std::string& params) noexcept
{
    receivedEventCount_[name]++;
}

void Tester::run() noexcept
{
    for (const auto& expectation : config_.getExpectations()) {
        resetCounters();

        const auto response = server_.runCmd(expectation.name_, expectation.request_.dump());

        // wait 100ms for upcoming events
        using namespace std::chrono_literals;
        std::this_thread::sleep_for(100ms);

        const auto report = createReport(response, expectation);

        if (report.pass) {
            std::cout << "[PASSED] run command: " << expectation.name_ << std::endl;
            continue;
        }

        std::cout << "[FAILED] run command: " << expectation.name_ << std::endl;
        if (report.missedCommands.empty() && report.missedEvents.empty()) {
            std::cout << boost::format(R"([REASON] unexpected response: "%1%"(real) vs %2%(expect))")
                         % response % detail::toString(*expectation.response_)
                      << std::endl;
        } else if (!report.missedCommands.empty()) {
            std::cout << boost::format("[REASON] missed commands: [%1%]")
                         % boost::algorithm::join(report.missedCommands, ", ")
                      << std::endl;
        } else {
            std::cout << boost::format("[REASON] missed events: [%1%]")
                         % boost::algorithm::join(report.missedEvents, ", ")
                      << std::endl;
        }
    }
}

} }
